
import os
from xepmts_server.secrets import MONGO_PASSWORD

VERSIONS = ['v1', 'v2', 'sc']

SERVER_SCHEMA = {
    'name': { 'type': 'string', 'nullable': False, 'unique': True},
    'url': { 'type': 'string', 'nullable': False, 'unique': True},

}

RESOURCE_SCHEMA = {
 'name': { 'type': 'string', 'nullable': False, 'unique': True},
 'url': { 'type': 'string', 'nullable': False, 'unique': True},
 'resource_methods': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'public_methods': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'item_methods': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'public_item_methods': { 'type': 'list', 'schema': {'type': 'string'}, },
 'allowed_roles': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'allowed_read_roles': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'allowed_write_roles': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'cache_control': { 'type': 'string', 'default': ''},
 'cache_expires': { 'type': 'integer', 'default': 0 },
 'id_field': { 'type': 'string', 'default': '_id'},
 'item_lookup_field': { 'type': 'string', 'default': '_id'},
 'item_url': { 'type': 'string', 'default': 'regex("[a-f0-9]{24}")'},
 'resource_title': { 'type': 'string', 'required': True},
 'item_title': { 'type': 'string', 'required': True},
 'item_lookup': { 'type': 'boolean', 'default': True},
 'allowed_item_roles': { 'type': 'list', 'schema': {'type': 'string'}},
 'allowed_item_read_roles': { 'type': 'list', 'schema': {'type': 'string'}},
 'allowed_item_write_roles': { 'type': 'list', 'schema': {'type': 'string'}},
 'allowed_filters': { 'type': 'list', 'schema': {'type': 'string'}, 'default': ['*']},
 'sorting': { 'type': 'boolean', 'default': True},
 'embedding': { 'type': 'boolean', 'default': True},
 'embedded_fields': { 'type': 'list', 'schema': {'type': 'string'}, 'default': []},
 'pagination': { 'type': 'boolean', 'default': True},
 'projection': { 'type': 'boolean', 'default': True},
 'versioning': { 'type': 'boolean', 'default': False},
 'soft_delete': { 'type': 'boolean', 'default': False},
 'bulk_enabled': { 'type': 'boolean', 'default': True},
 'internal_resource': { 'type': 'boolean', 'default': False},
 'etag_ignore_fields': { 'type': 'list', 'schema': {'type': 'string'}, 'nullable': True},
 'auth_field': { 'type': 'string', 'nullable': True},
 'allow_unknown': { 'type': 'boolean', 'default': False},
 'extra_response_fields': { 'type': 'list', 'schema': {'type': 'string'}, 'default': []},
 'mongo_query_whitelist': { 'type': 'list', 'schema': {'type': 'string'}, 'default': []},
 'mongo_write_concern':  { 'type': 'dict', 'default': {'w': 1}},
 'mongo_indexes': { 'type': 'dict', 'default': {}},
 'mongo_prefix': {'type': 'string', 'nullable': True},
 'hateoas': { 'type': 'boolean', 'default': True},
 'merge_nested_documents': { 'type': 'boolean', 'default': True},
 'normalize_dotted_fields': { 'type': 'boolean', 'default': True},
 'normalize_on_patch': { 'type': 'boolean', 'default': True},
 'schema': {'type': 'dict', 'required': True },
 'datasource': {
     'type': 'dict',
     'schema': {
          'source': {'type': 'string', },
          'filter': {'type': 'dict', 'nullable': True},
          'default_sort': {'type': 'string', 'nullable': True },
          'projection': {'type': 'dict', 'nullable': True},
          'aggregation': {'type': 'dict', 'nullable': True},
     }
               },
'additional_lookup': {'oneof': [{'type': 'string'}, {'type': 'dict'}], 'nullable': True},

}


DOMAIN = {
    v: {
        'schema': RESOURCE_SCHEMA,
        'datasource': {'source': f'{v}_endpoints'},
        'public_methods': ['GET'],
        'allowed_item_write_roles' : ['admin'],
        'allow_unknown': True,
        'additional_lookup': {
            'url': 'regex("[\w]+")',
            'field': 'name'
            }}
    for v in VERSIONS}

DOMAIN['servers'] = {
    'schema': SERVER_SCHEMA,
    'allowed_item_write_roles' : ['admin'],
    'allow_unknown': True,
    'additional_lookup': {
        'url': 'regex("[\w]+")',
        'field': 'name'
            },
}    

URL_PREFIX = os.getenv("XEPMTS_URL_PREFIX", "")
API_VERSION = "admin"
RESOURCE_METHODS = ["GET", "POST"]
ITEM_METHODS = ["GET", "PUT", "PATCH", "DELETE"]
ALLOWED_READ_ROLES = ["admin", "superuser", "expert", "user", "read", "write"]
ALLOWED_WRITE_ROLES = ["admin", "superuser", "expert", "write"]
EMBEDDING = True
MEDIA_PATH = "files"
PAGINATION_LIMIT = 10000
SCHEMA_ENDPOINT = "schema"
IF_MATCH = True
ENFORCE_IF_MATCH = False
HATEOAS = True
VERSIONS = "_versions"
NORMALIZE_ON_PATCH = False
ALLOW_UNKNOWN = True
# ----------------- Mongo config ------------------------------------------ #
MONGO_HOST = os.getenv("XEPMTS_MONGO_HOST", "localhost")
MONGO_PORT = int(os.getenv("XEPMTS_MONGO_PORT", 27017))
MONGO_DBNAME = os.getenv("XEPMTS_MONGO_DB", "pmts")
MONGO_USERNAME = os.getenv("XEPMTS_MONGO_USER", "")
MONGO_AUTH_SOURCE = os.getenv("XEPMTS_MONGO_AUTH_SOURCE", MONGO_DBNAME)

X_HEADERS = ['Content-Type', 'If-Match', 'Authorization', 'X-HTTP-Method-Override']  # Needed for the "Try it out" buttons

JWT_AUDIENCES = os.getenv("XEPMTS_JWT_AUDIENCES", "https://api.pmts.xenonnt.org").split(",")
JWT_KEY_URL = os.getenv("XEPMTS_JWT_KEY_URL", "https://auth-dot-xenon-pmts.uc.r.appspot.com/.well-known/jwks.json")
JWT_SCOPE_CLAIM = os.getenv("XEPMTS_JWT_SCOPE_CLAIM","scope")
JWT_ROLES_CLAIM = os.getenv("XEPMTS_JWT_ROLES_CLAIM", "roles")
JWT_TTL = 3600



import xepmts_server
from werkzeug.wsgi import pop_path_info, peek_path_info
from flask import request

import json
import time

REFRESH_PERIOD = 3600


def clean_dict(d):
    return {k1: {k2:v2 for k2,v2 in v1.items() if is_jsonable(v2)} for k1,v1 in d.items()}

def add_server_spec_endpoint(app):
    @app.route(f'/{app.config["API_VERSION"]}/server-spec')
    def server_spec():
        spec = {
            'base_url': f'{request.base_url}/{app.config["API_VERSION"]}/', 
            'endpoints': clean_dict(app.config['DOMAIN'])
        }
        return spec
    return app
    
class PathMakerDispatcher:
    def __init__(self, base_app, static_apps, app_configs):
        self.base_app = base_app
        self.static_apps = static_apps
        self.app_configs = app_configs
        self.dynamic_apps = {v: (time.time(), self.make_app(v)) for v in app_configs}
    
    def make_app(self, v): 
        return xepmts_server.get_server(v, **self.app_configs[v])

    def get_app(self, v):
        if v in self.static_apps:
            return self.static_apps[v]

        if v in self.dynamic_apps:
            t,app = self.dynamic_apps[v]
            if REFRESH_PERIOD<(time.time()-t):
                del app
                self.dynamic_apps[v] = t,app = time.time(), self.make_app(v)
            return app
        return self.base_app

    def __call__(self, environ, start_response):
        app = self.get_app(peek_path_info(environ))
        return app(environ, start_response)

class PathDispatcher:

    def __init__(self, base_app, app_dict):
        self.base_app = base_app
        self.apps = app_dict
        

    def __call__(self, environ, start_response):
        app = self.apps.get(peek_path_info(environ), self.base_app)
        return app(environ, start_response)

def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False



"""Console script for xepmts_endpoints."""
import sys

import click
from xepmts_server import make_app
from werkzeug.serving import run_simple
from flask.cli import with_appcontext


@click.group()
def main():
    """Console script for xepmts_endpoints."""
    click.echo("Replace this message by putting your code into "
               "xepmts_endpoints.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


@main.command()
@with_appcontext
def run(address='0.0.0.0', port=5000):
    app = make_app()
    run_simple(address, port, app,
                use_reloader=False, use_debugger=False, use_evalex=False)

@main.command()
@with_appcontext
def debug(address='0.0.0.0', port=5000):
    app = make_app(debug=True)
    run_simple(address, port, app,
                use_reloader=True, use_debugger=True, use_evalex=True)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import os

from flask import Flask


from threading import Lock
from werkzeug.wsgi import pop_path_info, peek_path_info
from werkzeug.serving import run_simple as _run_simple

from .v1.app import make_app as make_v1_app
from .v1.auth import XenonTokenAuth
from .v2.app import make_app as make_v2_app
from .admin.app import make_app as make_admin_app

from .utils import PathDispatcher, PathMakerDispatcher, add_server_spec_endpoint
import xepmts_server
from eve_jwt import JWTAuth

APP_MAKERS = {
    'v1': lambda: make_v1_app(auth=XenonTokenAuth, swagger=True),
    'v2': lambda: make_v2_app(auth=JWTAuth, swagger=True)
}

PRODUCTION_CONFIGS = {
    'v1': dict(auth=XenonTokenAuth, swagger=True),
    'v2': dict(auth=JWTAuth, swagger=True)
}

DEBUG_CONFIGS = {
    'v1': dict(swagger=True),
    'v2': dict(swagger=True),
}

def create_app():
    from eve_jwt import JWTAuth
    from flask_swagger_ui import get_swaggerui_blueprint
    
    v1 = make_v1_app(auth=XenonTokenAuth, swagger=True)
    v2 = make_v2_app(auth=JWTAuth, swagger=True)
    
    app_versions = {
        "v1": v1, 
        "v2": v2,
        }

    app = Flask(__name__)
    
    # @app.route("/")
    # def hello():
    #     return "You have reached the PMT db."

    app.config['SWAGGER_INFO'] = {
            'title': 'XENON PMT API',
            'version': '1.0',
            'description': 'API for the XENON PMT database',
            'termsOfService': 'https://opensource.org/ToS',
            'contact': {
                'name': 'Yossi Mosbacher',
                'url': 'https://pmts.xenonnt.org',
                "email": "joe.mosbacher@gmail.com"
            },

            'license': {
                'name': 'BSD',
                'url': 'https://github.com/nicolaiarocci/eve-swagger/blob/master/LICENSE',
            
            },
            'schemes': ['http', 'https'],

        }
    config = {
        'app_name': "PMT Database API",
        "urls": [{"name": f"Xenon PMT Database {v.capitalize()}", "url": f"/{v}/api-docs" } for v in app_versions]
    }
    API_URL = '/v2/api-docs'
    SWAGGER_URL = os.getenv('SWAGGER_URL_PREFIX', '')
    SWAGGERUI_BLUEPRINT = get_swaggerui_blueprint(
        SWAGGER_URL,
        API_URL,
        config=config,
    )
    app.register_blueprint(SWAGGERUI_BLUEPRINT, url_prefix=SWAGGER_URL)
    application = PathDispatcher(app,
                         app_versions)

    return application

def settings_dict(module):
    return {k: getattr(module, k) for k in dir(module) if k.isupper()}

def make_app(debug=False, overides={}, export_metrics=True, healthcheck=True):
    from eve_jwt import JWTAuth
    from flask_swagger_ui import get_swaggerui_blueprint
    
    # if versions is None:
    #     versions = xepmts_server.VERSIONS
    admin_auth = JWTAuth
    if debug:
        admin_auth = None
    admin = make_admin_app(auth=admin_auth, swagger=True)
    static_apps = {"admin": admin}
    if debug:
        configs = DEBUG_CONFIGS
    else:
        configs = PRODUCTION_CONFIGS
    
    app_configs = {}
    with admin.app_context():
        for version, config_kwargs in configs.items():
            kwargs = dict(config_kwargs)
            kwargs.update(overides)
            vmodule = getattr(xepmts_server, version)
            settings = kwargs.get('settings', settings_dict(vmodule.settings))
       
            endpoints = admin.data.driver.db[f'{version}_endpoints'].find()
            endpoints = {endpoint.pop('name'): endpoint for endpoint in endpoints}
            if endpoints:
                print(f'endpoints for version {version} taken from database.')
                settings['DOMAIN'] = endpoints
            kwargs['settings'] = settings
            kwargs['healthcheck'] = healthcheck
            kwargs['export_metrics'] = export_metrics
            app_configs[version] = kwargs
 

    app = Flask(__name__)

    app.config['SWAGGER_INFO'] = {
            'title': 'XENON PMT API',
            'version': '1.0',
            'description': 'API for the XENON PMT database',
            'termsOfService': 'https://opensource.org/ToS',
            'contact': {
                'name': 'Yossi Mosbacher',
                'url': 'https://pmts.xenonnt.org',
                "email": "joe.mosbacher@gmail.com"
            },

            'license': {
                'name': 'BSD',
                'url': 'https://github.com/nicolaiarocci/eve-swagger/blob/master/LICENSE',
            
            },
            'schemes': ['http', 'https'],

        }
    swagger_config = {
        'app_name': "PMT Database API",
        "urls": [{"name": f"Xenon PMT Database {v.capitalize()}", "url": f"/{v}/api-docs" } for v in list(app_configs)+list(static_apps)]
    }
    API_URL = '/v2/api-docs'
    SWAGGER_URL = os.getenv('SWAGGER_URL_PREFIX', '')
    SWAGGERUI_BLUEPRINT = get_swaggerui_blueprint(
        SWAGGER_URL,
        API_URL,
        config=swagger_config,
    )
    app.register_blueprint(SWAGGERUI_BLUEPRINT, url_prefix=SWAGGER_URL)
    if export_metrics:
        from prometheus_flask_exporter import PrometheusMetrics
        PrometheusMetrics(app)

    application = PathMakerDispatcher(app,
                          static_apps=static_apps,
                          app_configs=app_configs)
    return application


def run_simple(address='0.0.0.0', port=5000, debug=True, reload=True, evalex=True):
    app = make_app(debug=debug)
    _run_simple(address, port, app,
                use_reloader=debug, use_debugger=reload, use_evalex=evalex)

if __name__ == '__main__':
    run_simple()
    